/*****************************************************************************\
|* 
|* C++ visualisation app for the IP space we have mapped to cities.
|* 
|* Author						: SjG
|* First created				: Sat Nov  8 17:43:07 GMT 2003
|* Modification history			: <none>
|* Tabstops						: 4
|* Licence						: GPL, See COPYING for details
|*
|*    This program is free software; you can redistribute it and/or modify
|*    it under the terms of the GNU General Public License as published by
|*    the Free Software Foundation; either version 2 of the License, or
|*    (at your option) any later version.
|*
|*    This program is distributed in the hope that it will be useful,
|*    but WITHOUT ANY WARRANTY; without even the implied warranty of
|*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*    GNU General Public License for more details.
|*
|*    You should have received a copy of the GNU General Public License
|*    along with this program; if not, write to the Free Software
|*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
|*
|*
\*****************************************************************************/

#include "cubeFetcher.h"

#include <cstdarg>
#include <cstring>
#include <cstdlib>


namespace CubeFetcher {


using namespace std;

int readString(TCPsocket sock, char *ptr, int maxlen);
int readN(TCPsocket sock, char *ptr, int nbytes);
int writeString(TCPsocket sock, int *, char *format, ...);
void lcase(char *str);



/*****************************************************************************\
|* Method - fetch the cubes and create new Cube entries in the array
\*****************************************************************************/
void CubeFetcher::fetch (list<Cube>& cubes, char *url)
{

        /*************************************************************************\
        |* Parse the URL. This isn't too robust - it must be http://<site><path>
        \*************************************************************************/
        if (url == NULL || strncmp(url, "http://", 7) != 0) {
                fprintf(stderr, "Cannot parse url '%s'\n", url);
                return;
        }
	
        char *host 	= strdup(url+7);
        char *slash	= strchr(host, '/');
        char *path	= strdup(slash);
        *slash		= '\0';

        /*************************************************************************\
        |* Resolve the hostname and open the socket on port 80
        \*************************************************************************/
        IPaddress ip;
        if(SDLNet_ResolveHost(&ip, host, 80) == -1) {
                fprintf(stderr, "SDLNet_ResolveHost: %s\n", SDLNet_GetError());
                return;
        }

        TCPsocket s = SDLNet_TCP_Open(&ip);
        if(!s) {
                printf("SDLNet_TCP_Open: %s\n", SDLNet_GetError());
                exit(2);
        }

        /*************************************************************************\
        |* Write the request
        \*************************************************************************/
        int len;
        int n = writeString(s, &len,  "GET %s HTTP/1.0\r\nHost: %s\r\n\r\n",
                        path, host);
	if (n != len) {
                printf("SDLNet_TCP_Send: %s\n", SDLNet_GetError());
                SDLNet_TCP_Close(s);
                return;
        }

	/*************************************************************************\
	|* Read the Headers and get the content length
	\*************************************************************************/
        char hdr[1024];
        strcpy(hdr, "dummy value");
        int size = -1;
        while (strlen(hdr) > 2) {
                readString(s, hdr, 1024);
                lcase(hdr);
                if (strncmp(hdr, "content-length:", 15) == 0)
                        size = atoi(hdr+16);
        }

        /*************************************************************************\
        |* Read the payload - 'size' units of 5 bytes
        \*************************************************************************/
        size /= 4;
        Uint8 clump[4];

        for (int i=0; i<size; i++) {
                readN(s, (char *)clump, 4);

                float xx = clump[0] / 255.f;
                float yy = clump[1] / 255.f;
                float zz = clump[2] / 255.f;
                float sz = 1.f/255.f;

                Uint8 R = (clump[3] & 0xe0);
                Uint8 G = (clump[3] & 0x1C) << 3;
                Uint8 B = (clump[3] & 0x03) << 6;

                cubes.push_back(Cube (xx, yy, zz, sz, R, G, B));
        }

        /*************************************************************************\
        |* Close it down and tidy up
        \*************************************************************************/
        SDLNet_TCP_Close(s);
        free(host);
        free(path);
}


/*****************************************************************************\
|* write a string over a socket
\*****************************************************************************/
int writeString(TCPsocket sock, int *len, char *format, ...) {
        char buf[2048];
        va_list list;

        memset(buf, 0, 2048);
        va_start (list, format);
        vsprintf (buf, format, list);
        va_end (list);

        if (buf[strlen(buf)-1] != '\n')
                strcat(buf, "\n");
        *len = strlen(buf);
        //fprintf(stderr, "Writing [%s]\n", buf);

        return SDLNet_TCP_Send(sock, buf, *len);
}

/*****************************************************************************\
|* read N bytes from a socket
\*****************************************************************************/
int readN(TCPsocket sock, char *ptr, int nbytes) {
        int nleft, nread;
        int flag;

        nleft = nbytes;
        while (nleft > 0) {
                /* -------------------------------------------------------------------
                // First apply a timeout to the read operation. If we don't get a read
                // back within the timeout period, we know the connection has somehow
                // become broken.
                //
                // We have to re-initialise all the fields on every call because Unix
                // (Linux especially) likes to alter fields on return
                // -----------------------------------------------------------------*/
                SDLNet_SocketSet set = SDLNet_AllocSocketSet(1);
                SDLNet_AddSocket(set, (SDLNet_GenericSocket)sock);
                flag = SDLNet_CheckSockets(set, 120000);

                if (flag) {
                        nread = SDLNet_TCP_Recv(sock, ptr, nleft);
                        if (nread <= 0)
                                return (nread);         /* error */

                        nleft -= nread;
                        ptr   += nread;
                } else {
                        // ---------------------------------------------------------------
                        // This is a timeout on the socket read
                        // ---------------------------------------------------------------
                        SDLNet_FreeSocketSet(set);
                        return -1;
                }

                SDLNet_FreeSocketSet(set);
        }
        return (nbytes - nleft);
}

/*****************************************************************************\
|* read a string from a socket
\*****************************************************************************/
int readString(TCPsocket sock, char *ptr, int maxlen) {
        int n, rc;
        char c;

        for (n = 1; n < maxlen; n++) {
                if ( (rc = SDLNet_TCP_Recv(sock, &c, 1)) == 1) {
                        *ptr++ = c;
                        if (c == '\n')
                                break;
                }
                else if (rc == 0) {
                        if (n == 1)
                                return(0);      /* EOF, no data read */
                        else
                                break;          /* EOF, some data was read */
                } else
                        return(-1);     		/* error */
        }

        *ptr = 0;
        return(n);
}

/*****************************************************************************\
|* convert a string to lowercase
\*****************************************************************************/
void lcase(char *str) {
        int size = strlen(str);

        for (int i=0; i<size; i++)
                if (str[i] >= 'A' && str[i] <= 'a')
                        str[i] += 'a' - 'A';
}
	
	
} // namespace
